from dataclasses import dataclass

from artifact_experiment.spi.resources import (
    CacheCallbackResources,
    CallbackResources,
    TrackingCallbackResources,
)


@dataclass(frozen=True)
class PeriodicCallbackResources(CallbackResources):
    step: int


@dataclass(frozen=True)
class PeriodicCacheCallbackResources(CacheCallbackResources):
    step: int


@dataclass(frozen=True)
class PeriodicTrackingCallbackResources(TrackingCallbackResources):
    step: int



from artifact_torch._base.components.routines.artifact import (
    ArtifactRoutine,
    ArtifactRoutineData,
    ArtifactRoutineHyperparams,
)
from artifact_torch._base.experiment.experiment import Experiment

__all__ = ["ArtifactRoutine", "ArtifactRoutineData", "ArtifactRoutineHyperparams", "Experiment"]



from abc import ABC, abstractmethod
from typing import Any, Generic, Mapping, Optional, Type, TypeVar

import pandas as pd
from artifact_core.spi.resources import ResourceSpecProtocol
from artifact_experiment.tracking import DataSplit
from artifact_experiment.tracking.spi import FileWriter, TrackingClient, TrackingQueue

from artifact_torch._base.components.routines.artifact import ArtifactRoutine, ArtifactRoutineData
from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine
from artifact_torch._base.data.data_loader import DataLoader
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.trainer import Trainer

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
ArtifactRoutineDataTContr = TypeVar(
    "ArtifactRoutineDataTContr", bound=ArtifactRoutineData, contravariant=True
)
ResourceSpecProtocolTContr = TypeVar(
    "ResourceSpecProtocolTContr", bound=ResourceSpecProtocol, contravariant=True
)
ExperimentT = TypeVar("ExperimentT", bound="Experiment")


class Experiment(
    ABC,
    Generic[
        ModelTContr,
        ModelInputTContr,
        ModelOutputTContr,
        ArtifactRoutineDataTContr,
        ResourceSpecProtocolTContr,
    ],
):
    def __init__(
        self, trainer: Trainer[ModelTContr, ModelInputTContr, ModelOutputTContr, Any, Any]
    ):
        self._trainer = trainer

    @classmethod
    def build(
        cls: Type[ExperimentT],
        model: ModelTContr,
        data_loaders: Mapping[DataSplit, DataLoader[ModelInputTContr]],
        artifact_routine_data: Mapping[DataSplit, ArtifactRoutineDataTContr],
        artifact_routine_data_spec: Optional[ResourceSpecProtocolTContr] = None,
        tracking_client: Optional[TrackingClient[Any]] = None,
    ) -> ExperimentT:
        assert DataSplit.TRAIN in data_loaders, "Training data not provided."
        tracking_queue = cls._get_tracking_queue(tracking_client=tracking_client)
        file_writer = cls._get_file_writer(tracking_client=tracking_client)
        train_diagnostics_routine = cls._build_train_diagnostics_routine(
            tracking_queue=tracking_queue
        )
        loader_routine = cls._build_loader_routine(
            data_loaders=data_loaders, tracking_queue=tracking_queue
        )
        artifact_routine = cls._build_artifact_routine(
            data=artifact_routine_data,
            data_spec=artifact_routine_data_spec,
            tracking_queue=tracking_queue,
        )
        trainer = cls._build_trainer(
            model=model,
            train_loader=data_loaders[DataSplit.TRAIN],
            train_diagnostics_routine=train_diagnostics_routine,
            loader_routine=loader_routine,
            artifact_routine=artifact_routine,
            file_writer=file_writer,
        )
        experiment = cls(trainer=trainer)
        return experiment

    @property
    def epoch_scores(self) -> pd.DataFrame:
        return self._trainer.epoch_scores

    @classmethod
    @abstractmethod
    def _get_trainer(
        cls,
    ) -> Type[Trainer[ModelTContr, ModelInputTContr, ModelOutputTContr, Any, Any]]: ...

    @classmethod
    @abstractmethod
    def _get_train_diagnostics_routine(
        cls,
    ) -> Optional[
        Type[TrainDiagnosticsRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_loader_routine(
        cls,
    ) -> Optional[Type[DataLoaderRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]]]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_routine(
        cls,
    ) -> Optional[
        Type[
            ArtifactRoutine[
                ModelTContr, ArtifactRoutineDataTContr, Any, Any, ResourceSpecProtocolTContr
            ]
        ]
    ]: ...

    def run(self):
        self._trainer.train()

    @classmethod
    def _build_trainer(
        cls,
        model: ModelTContr,
        train_loader: DataLoader[ModelInputTContr],
        train_diagnostics_routine: Optional[
            TrainDiagnosticsRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ],
        loader_routine: Optional[
            DataLoaderRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ],
        artifact_routine: Optional[
            ArtifactRoutine[
                ModelTContr, ArtifactRoutineDataTContr, Any, Any, ResourceSpecProtocolTContr
            ]
        ],
        file_writer: Optional[FileWriter],
    ) -> Trainer[ModelTContr, ModelInputTContr, ModelOutputTContr, Any, Any]:
        trainer_class = cls._get_trainer()
        trainer = trainer_class.build(
            model=model,
            train_loader=train_loader,
            train_diagnostics_routine=train_diagnostics_routine,
            loader_routine=loader_routine,
            artifact_routine=artifact_routine,
            file_writer=file_writer,
        )
        return trainer

    @classmethod
    def _build_train_diagnostics_routine(
        cls, tracking_queue: Optional[TrackingQueue]
    ) -> Optional[TrainDiagnosticsRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]]:
        routine = None
        routine_class = cls._get_train_diagnostics_routine()
        if routine_class is not None:
            routine = routine_class.build(tracking_queue=tracking_queue)
        return routine

    @classmethod
    def _build_loader_routine(
        cls,
        data_loaders: Mapping[DataSplit, DataLoader[ModelInputTContr]],
        tracking_queue: Optional[TrackingQueue],
    ) -> Optional[DataLoaderRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]]:
        routine = None
        routine_class = cls._get_loader_routine()
        if routine_class is not None:
            routine = routine_class.build(data_loaders=data_loaders, tracking_queue=tracking_queue)
        return routine

    @classmethod
    def _build_artifact_routine(
        cls,
        data: Mapping[DataSplit, ArtifactRoutineDataTContr],
        data_spec: Optional[ResourceSpecProtocolTContr],
        tracking_queue: Optional[TrackingQueue],
    ) -> Optional[
        ArtifactRoutine[
            ModelTContr, ArtifactRoutineDataTContr, Any, Any, ResourceSpecProtocolTContr
        ]
    ]:
        routine = None
        routine_class = cls._get_artifact_routine()
        if routine_class is not None and data_spec is not None:
            routine = routine_class.build(
                data=data, data_spec=data_spec, tracking_queue=tracking_queue
            )
        return routine

    @staticmethod
    def _get_tracking_queue(tracking_client: Optional[TrackingClient]) -> Optional[TrackingQueue]:
        return tracking_client.queue if tracking_client is not None else None

    @staticmethod
    def _get_file_writer(tracking_client: Optional[TrackingClient]) -> Optional[FileWriter]:
        return tracking_client.file_writer if tracking_client is not None else None



from abc import abstractmethod
from typing import Any, Generic, Optional, Type, TypeVar

from artifact_core.table_comparison.spi import TabularDataSpecProtocol

from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine
from artifact_torch._base.experiment.experiment import Experiment
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.trainer import Trainer
from artifact_torch._domains.generation.model import GenerationParams
from artifact_torch.table_comparison._model import TableSynthesizer
from artifact_torch.table_comparison._routine import (
    TableComparisonRoutine,
    TableComparisonRoutineData,
)

TableSynthesizerTContr = TypeVar(
    "TableSynthesizerTContr", bound=TableSynthesizer[Any, Any, Any], contravariant=True
)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
GenerationParamsTContr = TypeVar(
    "GenerationParamsTContr", bound=GenerationParams, contravariant=True
)
TabularSynthesisExperimentT = TypeVar(
    "TabularSynthesisExperimentT", bound="TabularSynthesisExperiment"
)


class TabularSynthesisExperiment(
    Experiment[
        TableSynthesizerTContr,
        ModelInputTContr,
        ModelOutputTContr,
        TableComparisonRoutineData,
        TabularDataSpecProtocol,
    ],
    Generic[TableSynthesizerTContr, ModelInputTContr, ModelOutputTContr, GenerationParamsTContr],
):
    @classmethod
    @abstractmethod
    def _get_trainer(
        cls,
    ) -> Type[
        Trainer[
            TableSynthesizerTContr,
            ModelInputTContr,
            ModelOutputTContr,
            Any,
            Any,
        ]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_train_diagnostics_routine(
        cls,
    ) -> Optional[
        Type[TrainDiagnosticsRoutine[TableSynthesizerTContr, ModelInputTContr, ModelOutputTContr]]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_loader_routine(
        cls,
    ) -> Optional[
        Type[DataLoaderRoutine[TableSynthesizerTContr, ModelInputTContr, ModelOutputTContr]]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_routine(
        cls,
    ) -> Optional[Type[TableComparisonRoutine[GenerationParamsTContr]]]: ...



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Sequence, Type, TypeVar

from artifact_core.typing import ArtifactResult
from artifact_experiment.spi.plans import CallbackExecutionPlan, PlanBuildContext

from artifact_torch._base.components.callbacks.backward_hook import (
    BackwardHookArrayCallback,
    BackwardHookArrayCollectionCallback,
    BackwardHookCallback,
    BackwardHookPlotCallback,
    BackwardHookPlotCollectionCallback,
    BackwardHookScoreCallback,
    BackwardHookScoreCollectionCallback,
)
from artifact_torch._base.components.handler_suites.hook import HookCallbackHandlerSuite
from artifact_torch._base.components.resources.hook import HookCallbackResources
from artifact_torch._base.model.base import Model


@dataclass(frozen=True)
class BackwardHookPlanBuildContext(PlanBuildContext):
    pass


ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
BackwardHookPlanT = TypeVar("BackwardHookPlanT", bound="BackwardHookPlan[Any]")


class BackwardHookPlan(
    CallbackExecutionPlan[
        HookCallbackHandlerSuite[ModelTContr],
        BackwardHookCallback[ModelTContr, ArtifactResult, Any],
        HookCallbackResources[ModelTContr],
        BackwardHookPlanBuildContext,
    ],
    Generic[ModelTContr],
):
    @classmethod
    @abstractmethod
    def _get_score_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookScoreCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookArrayCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookPlotCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_score_collection_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookScoreCollectionCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_collection_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookArrayCollectionCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_collection_callbacks(
        cls, context: BackwardHookPlanBuildContext
    ) -> Sequence[BackwardHookPlotCollectionCallback[ModelTContr]]: ...

    def attach(self, resources: HookCallbackResources[ModelTContr]) -> bool:
        return self._handler_suite.attach(resources=resources)

    @classmethod
    def _get_handler_suite(cls) -> Type[HookCallbackHandlerSuite[ModelTContr]]:
        return HookCallbackHandlerSuite



from abc import abstractmethod
from collections.abc import Sequence
from typing import Dict, Generic, List, Optional, Tuple, TypeVar

import torch
import torch.nn as nn
from artifact_core.typing import (
    Array,
    ArrayCollection,
    ArtifactResult,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)
from torch.utils.hooks import RemovableHandle

from artifact_torch._base.components.callbacks.hook import HookCallback
from artifact_torch._base.model.base import Model

ModelTContr = TypeVar("ModelTContr", bound=Model, contravariant=True)
CacheDataTCov = TypeVar("CacheDataTCov", bound=ArtifactResult, covariant=True)
HookResultT = TypeVar("HookResultT")


class BackwardHookCallback(
    HookCallback[ModelTContr, CacheDataTCov, HookResultT],
    Generic[ModelTContr, CacheDataTCov, HookResultT],
):
    @classmethod
    @abstractmethod
    def _get_base_key(cls) -> str: ...

    @classmethod
    @abstractmethod
    def _get_layers(cls, model: ModelTContr) -> Sequence[nn.Module]: ...

    @classmethod
    @abstractmethod
    def _hook(
        cls,
        module: nn.Module,
        grad_input: Tuple[Optional[torch.Tensor], ...],
        grad_output: Tuple[Optional[torch.Tensor], ...],
    ) -> Optional[HookResultT]: ...

    @classmethod
    @abstractmethod
    def _aggregate(cls, hook_results: Dict[str, List[HookResultT]]) -> CacheDataTCov: ...

    @classmethod
    def _attach(
        cls, model: ModelTContr, sink: Dict[str, List[HookResultT]], handles: List[RemovableHandle]
    ):
        def _wrapped_hook(
            module: nn.Module,
            grad_input: Tuple[Optional[torch.Tensor], ...],
            grad_output: Tuple[Optional[torch.Tensor], ...],
        ):
            hook_return_value = cls._hook(module, grad_input, grad_output)
            if hook_return_value is not None:
                module_name = f"{module.__class__.__name__}"
                if module_name not in sink:
                    sink[module_name] = []
                sink[module_name].append(hook_return_value)

        for module in cls._get_layers(model):
            handle = module.register_full_backward_hook(_wrapped_hook)  # type: ignore[reportArgumentType]
            handles.append(handle)


BackwardHookScoreCallback = BackwardHookCallback[ModelTContr, Score, Dict[str, torch.Tensor]]

BackwardHookArrayCallback = BackwardHookCallback[ModelTContr, Array, Dict[str, torch.Tensor]]


BackwardHookPlotCallback = BackwardHookCallback[ModelTContr, Plot, Dict[str, torch.Tensor]]


BackwardHookScoreCollectionCallback = BackwardHookCallback[
    ModelTContr, ScoreCollection, Dict[str, torch.Tensor]
]


BackwardHookArrayCollectionCallback = BackwardHookCallback[
    ModelTContr, ArrayCollection, Dict[str, torch.Tensor]
]


BackwardHookPlotCollectionCallback = BackwardHookCallback[
    ModelTContr, PlotCollection, Dict[str, torch.Tensor]
]



from artifact_torch._base.data.data_loader import DataLoader
from artifact_torch._base.data.dataset import Dataset
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.trainer import Trainer

__all__ = [
    "DataLoader",
    "Dataset",
    "Model",
    "ModelInput",
    "ModelOutput",
    "Trainer",
]



from abc import abstractmethod
from typing import Any, Generic, List, Mapping, Optional, Type, TypeVar

import torch
from artifact_experiment.tracking import DataSplit
from artifact_experiment.tracking.spi import TrackingQueue
from tqdm import tqdm

from artifact_torch._base.components.plans.forward_hook import (
    ForwardHookPlan,
    ForwardHookPlanBuildContext,
)
from artifact_torch._base.components.plans.model_io import ModelIOPlan, ModelIOPlanBuildContext
from artifact_torch._base.components.resources.hook import HookCallbackResources
from artifact_torch._base.components.routines.routine import PlanExecutionRoutine, RoutineResources
from artifact_torch._base.data.data_loader import DataLoader
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._utils.collections.key_selector import KeySelector

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
DataLoaderRoutineT = TypeVar("DataLoaderRoutineT", bound="DataLoaderRoutine[Any, Any, Any]")


class DataLoaderRoutine(
    PlanExecutionRoutine[ModelTContr], Generic[ModelTContr, ModelInputTContr, ModelOutputTContr]
):
    _verbose = True
    _progressbar_message = "Processing Data Loader"

    def __init__(
        self,
        model_io_plans: Mapping[DataSplit, ModelIOPlan[ModelInputTContr, ModelOutputTContr]],
        forward_hook_plans: Mapping[DataSplit, ForwardHookPlan[ModelTContr]],
        data_loaders: Mapping[DataSplit, DataLoader[ModelInputTContr]],
    ):
        self._data_loaders = data_loaders
        self._model_io_plans = KeySelector.restrict_to_keys(model_io_plans, keys_from=data_loaders)
        self._forward_hook_plans = KeySelector.restrict_to_keys(
            forward_hook_plans, keys_from=data_loaders
        )
        plans = list(self._model_io_plans.values()) + list(self._forward_hook_plans.values())
        super().__init__(plans=plans)

    @classmethod
    def build(
        cls: Type[DataLoaderRoutineT],
        data_loaders: Mapping[DataSplit, DataLoader[ModelInputTContr]],
        tracking_queue: Optional[TrackingQueue] = None,
    ) -> DataLoaderRoutineT:
        model_io_build_context = ModelIOPlanBuildContext(tracking_queue=tracking_queue)
        forward_hook_build_context = ForwardHookPlanBuildContext(tracking_queue=tracking_queue)
        model_io_plans = {
            data_split: plan
            for data_split in DataSplit
            if (
                plan := cls._build_model_io_plan(
                    data_split=data_split, context=model_io_build_context
                )
            )
            is not None
        }
        forward_hook_plans = {
            data_split: plan
            for data_split in DataSplit
            if (
                plan := cls._build_forward_hook_plan(
                    data_split=data_split, context=forward_hook_build_context
                )
            )
            is not None
        }
        routine = cls(
            model_io_plans=model_io_plans,
            forward_hook_plans=forward_hook_plans,
            data_loaders=data_loaders,
        )
        return routine

    @property
    def _data_splits(self) -> List[DataSplit]:
        return list(self._data_loaders.keys())

    @classmethod
    @abstractmethod
    def _get_model_io_plan(
        cls, data_split: DataSplit
    ) -> Optional[Type[ModelIOPlan[ModelInputTContr, ModelOutputTContr]]]: ...

    @classmethod
    @abstractmethod
    def _get_forward_hook_plan(
        cls, data_split: DataSplit
    ) -> Optional[Type[ForwardHookPlan[ModelTContr]]]: ...

    def execute(self, resources: RoutineResources[ModelTContr]):
        for data_split in self._data_splits:
            callback_resources = HookCallbackResources[ModelTContr](
                model=resources.model, step=resources.n_epochs_elapsed, data_split=data_split
            )
            data_loader = self._data_loaders[data_split]
            model_io_plan = self._model_io_plans.get(data_split)
            forward_hook_plan = self._forward_hook_plans.get(data_split)
            self._execute(
                callback_resources=callback_resources,
                model_io_plan=model_io_plan,
                forward_hook_plan=forward_hook_plan,
                data_loader=data_loader,
            )

    @classmethod
    def _execute(
        cls,
        callback_resources: HookCallbackResources[ModelTContr],
        model_io_plan: Optional[ModelIOPlan[ModelInputTContr, ModelOutputTContr]],
        forward_hook_plan: Optional[ForwardHookPlan[ModelTContr]],
        data_loader: DataLoader[ModelInputTContr],
    ):
        any_attached = cls._attach(
            callback_resources=callback_resources,
            model_io_plan=model_io_plan,
            forward_hook_plan=forward_hook_plan,
        )
        if any_attached:
            cls._process_data_loader(model=callback_resources.model, data_loader=data_loader)
            if model_io_plan is not None:
                model_io_plan.execute(resources=callback_resources)
            if forward_hook_plan is not None:
                forward_hook_plan.execute(resources=callback_resources)

    @staticmethod
    def _attach(
        callback_resources: HookCallbackResources[ModelTContr],
        model_io_plan: Optional[ModelIOPlan[ModelInputTContr, ModelOutputTContr]],
        forward_hook_plan: Optional[ForwardHookPlan[ModelTContr]],
    ) -> bool:
        any_attached = False
        if model_io_plan is not None:
            any_attached |= model_io_plan.attach(resources=callback_resources)
        if forward_hook_plan is not None:
            any_attached |= forward_hook_plan.attach(resources=callback_resources)
        return any_attached

    @classmethod
    def _process_data_loader(cls, model: ModelTContr, data_loader: DataLoader[ModelInputTContr]):
        model.eval()
        with torch.no_grad():
            for model_input in tqdm(
                data_loader,
                desc=cls._progressbar_message,
                disable=not cls._verbose,
                leave=False,
            ):
                _ = model(model_input)

    @classmethod
    def _build_model_io_plan(
        cls, data_split: DataSplit, context: ModelIOPlanBuildContext
    ) -> Optional[ModelIOPlan[ModelInputTContr, ModelOutputTContr]]:
        plan_class = cls._get_model_io_plan(data_split=data_split)
        if plan_class is not None:
            return plan_class.build(context=context)

    @classmethod
    def _build_forward_hook_plan(
        cls, data_split: DataSplit, context: ForwardHookPlanBuildContext
    ) -> Optional[ForwardHookPlan[ModelTContr]]:
        plan_class = cls._get_forward_hook_plan(data_split=data_split)
        if plan_class is not None:
            return plan_class.build(context=context)



from pathlib import Path
from typing import Any, Dict, Generic, Optional, TypeVar, Union

import torch
from torch import optim

from artifact_torch._base.model.base import Model

ModelTCov = TypeVar("ModelTCov", bound=Model[Any, Any], covariant=True)


class TrainingState(Generic[ModelTCov]):
    _model_state_checkpoint_key: str = "model_state"
    _optimizer_state_checkpoint_key: str = "optimizer_state"
    _scheduler_state_checkpoint_key: str = "scheduler_state"

    def __init__(
        self,
        model: ModelTCov,
        optimizer: optim.Optimizer,
        scheduler: Optional[optim.lr_scheduler._LRScheduler],
    ):
        self._model = model
        self._optimizer = optimizer
        self._scheduler = scheduler
        self._validate()

    @property
    def model(self) -> ModelTCov:
        return self._model

    @property
    def optimizer(self) -> optim.Optimizer:
        return self._optimizer

    @property
    def scheduler(self) -> Optional[optim.lr_scheduler._LRScheduler]:
        return self._scheduler

    def serialize(self) -> Dict[str, Any]:
        checkpoint = {}
        checkpoint[self._model_state_checkpoint_key] = self._model.state_dict()
        checkpoint[self._optimizer_state_checkpoint_key] = self._optimizer.state_dict()
        if self._scheduler is not None:
            checkpoint[self._scheduler_state_checkpoint_key] = self._scheduler.state_dict()
        return checkpoint

    def export(self, path: Union[Path, str]):
        path = self._validate_checkpoint_path(path=path)
        checkpoint = self.serialize()
        torch.save(checkpoint, path)

    def load_checkpoint(self, checkpoint: Dict[str, Any]):
        self._model.load_state_dict(checkpoint.get(self._model_state_checkpoint_key, {}))
        self._optimizer.load_state_dict(checkpoint.get(self._optimizer_state_checkpoint_key, {}))
        if self._scheduler is not None:
            self._scheduler.load_state_dict(
                checkpoint.get(self._scheduler_state_checkpoint_key, {})
            )

    def load(self, path: Union[Path, str], map_location: torch.device):
        checkpoint: Dict[str, Any] = torch.load(path, weights_only=True, map_location=map_location)
        self.load_checkpoint(checkpoint=checkpoint)

    def _validate(self):
        self._validate_optimizer(optimizer=self._optimizer, model=self._model)
        if self._scheduler is not None:
            self._validate_scheduler(
                scheduler=self._scheduler,
                optimizer=self._optimizer,
            )

    @staticmethod
    def _validate_optimizer(optimizer: optim.Optimizer, model: Model[Any, Any]):
        model_param_ids = {id(p) for p in model.parameters()}
        optimizer_param_ids = {id(p) for group in optimizer.param_groups for p in group["params"]}
        if not model_param_ids.issubset(optimizer_param_ids):
            raise ValueError("Optimizer not properly tied to model params.")

    @staticmethod
    def _validate_scheduler(scheduler: optim.lr_scheduler._LRScheduler, optimizer: optim.Optimizer):
        if scheduler.optimizer is not optimizer:
            raise ValueError("Scheduler not properly tied to optimizer.")

    @staticmethod
    def _validate_checkpoint_path(path: Union[str, Path]) -> Path:
        path = Path(path)
        if path.suffix != ".pth":
            raise ValueError(f"Invalid file extension: {path.suffix}. Expected '.pth'")
        return path



from abc import abstractmethod
from dataclasses import dataclass

from artifact_torch._base.components.model_tracking.model_tracker import (
    ModelTracker,
    ModelTrackingCriterion,
)


@dataclass
class SingleScoreCriterion(ModelTrackingCriterion):
    score: float


class SingleScoreTracker(ModelTracker[SingleScoreCriterion]):
    def __init__(self):
        self.reset()

    @property
    def best_score(self) -> float:
        return self._best_criterion.score

    @staticmethod
    @abstractmethod
    def _new_score_is_best(new_score: float, best_score: float) -> bool: ...

    @staticmethod
    @abstractmethod
    def _get_initial_best_score() -> float: ...

    @classmethod
    def _new_criterion_is_best(
        cls, new_criterion: SingleScoreCriterion, best_criterion: SingleScoreCriterion
    ) -> bool:
        new_criterion_is_best = cls._new_score_is_best(
            new_score=new_criterion.score, best_score=best_criterion.score
        )
        return new_criterion_is_best

    @classmethod
    def _get_initial_best_criterion(cls) -> SingleScoreCriterion:
        initial_best_score = cls._get_initial_best_score()
        initial_best_criterion = SingleScoreCriterion(score=initial_best_score)
        return initial_best_criterion



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, TypeVar

from artifact_torch._base.model.base import Model

ModelTrackingCriterionT = TypeVar("ModelTrackingCriterionT", bound="ModelTrackingCriterion")


@dataclass
class ModelTrackingCriterion: ...


@dataclass
class NoTrackingCriterion(ModelTrackingCriterion): ...


class ModelTracker(ABC, Generic[ModelTrackingCriterionT]):
    def __init__(self):
        self.reset()

    @property
    def best_model_state(self) -> Dict[str, Any]:
        return self._best_model_state

    @property
    def best_criterion(self) -> ModelTrackingCriterionT:
        return self._best_criterion

    @property
    def best_epoch(self) -> int:
        return self._best_epoch

    @classmethod
    @abstractmethod
    def _new_criterion_is_best(
        cls, new_criterion: ModelTrackingCriterionT, best_criterion: ModelTrackingCriterionT
    ) -> bool: ...

    @classmethod
    @abstractmethod
    def _get_initial_best_criterion(cls) -> ModelTrackingCriterionT: ...

    def update(self, model: Model[Any, Any], criterion: ModelTrackingCriterionT, epoch: int):
        if self._new_criterion_is_best(
            new_criterion=criterion, best_criterion=self._best_criterion
        ):
            self._best_score = criterion
            self._best_model_state = model.state_dict()
            self._best_epoch = epoch

    def reset(self):
        self._best_model_state = {}
        self._best_criterion = self._get_initial_best_criterion()
        self._best_epoch = 0



from typing import Any, Hashable, Mapping, Set, TypeVar

HashableT = TypeVar("HashableT", bound=Hashable)
ValueT = TypeVar("ValueT")


class KeySelector:
    @staticmethod
    def get_non_null_keys(d: Mapping[HashableT, Any]) -> Set[HashableT]:
        return {k for k, v in d.items() if v is not None}

    @staticmethod
    def get_common_non_null_keys(*dicts: Mapping[HashableT, Any]) -> Set[HashableT]:
        if not dicts:
            return set()
        common = set(dicts[0]).intersection(*dicts[1:])
        return {k for k in common if all(d.get(k) is not None for d in dicts)}

    @staticmethod
    def get_common_keys(*dicts: Mapping[HashableT, Any]) -> Set[HashableT]:
        return set(dicts[0]).intersection(*dicts[1:]) if dicts else set()

    @staticmethod
    def restrict_to_keys(
        d: Mapping[HashableT, ValueT],
        keys_from: Mapping[HashableT, Any],
    ) -> dict[HashableT, ValueT]:
        key_set = set(keys_from)
        return {k: v for k, v in d.items() if k in key_set}



from abc import abstractmethod
from typing import Any, Generic, Optional, Type, TypeVar

from artifact_experiment.tracking.spi import TrackingQueue

from artifact_torch._base.components.callbacks.hook import HookCallbackResources
from artifact_torch._base.components.plans.backward_hook import (
    BackwardHookPlan,
    BackwardHookPlanBuildContext,
)
from artifact_torch._base.components.plans.forward_hook import (
    ForwardHookPlan,
    ForwardHookPlanBuildContext,
)
from artifact_torch._base.components.plans.model_io import ModelIOPlan, ModelIOPlanBuildContext
from artifact_torch._base.components.routines.routine import PlanExecutionRoutine, RoutineResources
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
TrainDiagnosticsRoutineT = TypeVar(
    "TrainDiagnosticsRoutineT", bound="TrainDiagnosticsRoutine[Any, Any, Any]"
)


class TrainDiagnosticsRoutine(
    PlanExecutionRoutine[ModelTContr], Generic[ModelTContr, ModelInputTContr, ModelOutputTContr]
):
    _callback_trigger_identifier = "EPOCH"

    def __init__(
        self,
        model_io_plan: Optional[ModelIOPlan[ModelInputTContr, ModelOutputTContr]],
        forward_hook_plan: Optional[ForwardHookPlan[ModelTContr]],
        backward_hook_plan: Optional[BackwardHookPlan[ModelTContr]],
    ):
        self._model_io_plan = model_io_plan
        self._forward_hook_plan = forward_hook_plan
        self._backward_hook_plan = backward_hook_plan
        plans = [
            plan
            for plan in [model_io_plan, forward_hook_plan, backward_hook_plan]
            if plan is not None
        ]
        super().__init__(plans=plans)

    @classmethod
    def build(
        cls: Type[TrainDiagnosticsRoutineT], tracking_queue: Optional[TrackingQueue] = None
    ) -> TrainDiagnosticsRoutineT:
        model_io_build_context = ModelIOPlanBuildContext(tracking_queue=tracking_queue)
        forward_hook_build_context = ForwardHookPlanBuildContext(tracking_queue=tracking_queue)
        backward_hook_build_context = BackwardHookPlanBuildContext(tracking_queue=tracking_queue)
        model_io_plan = cls._build_model_io_plan(context=model_io_build_context)
        forward_hook_plan = cls._build_forward_hook_plan(context=forward_hook_build_context)
        backward_hook_plan = cls._build_backward_hook_plan(context=backward_hook_build_context)
        routine = cls(
            model_io_plan=model_io_plan,
            forward_hook_plan=forward_hook_plan,
            backward_hook_plan=backward_hook_plan,
        )
        return routine

    @classmethod
    @abstractmethod
    def _get_model_io_plan(
        cls,
    ) -> Optional[Type[ModelIOPlan[ModelInputTContr, ModelOutputTContr]]]: ...

    @classmethod
    @abstractmethod
    def _get_forward_hook_plan(cls) -> Optional[Type[ForwardHookPlan[ModelTContr]]]: ...

    @classmethod
    @abstractmethod
    def _get_backward_hook_plan(cls) -> Optional[Type[BackwardHookPlan[ModelTContr]]]: ...

    def execute(self, resources: RoutineResources[ModelTContr]):
        callback_resources = HookCallbackResources[ModelTContr](
            model=resources.model,
            step=resources.n_epochs_elapsed,
            trigger=self._callback_trigger_identifier,
        )
        if self._model_io_plan is not None:
            self._model_io_plan.execute(resources=callback_resources)
        if self._forward_hook_plan is not None:
            self._forward_hook_plan.execute(resources=callback_resources)
        if self._backward_hook_plan is not None:
            self._backward_hook_plan.execute(resources=callback_resources)

    def attach(self, model: ModelTContr, n_epochs_elapsed: int) -> bool:
        resources = HookCallbackResources[ModelTContr](model=model, step=n_epochs_elapsed)
        any_attached = False
        if self._model_io_plan is not None:
            any_attached |= self._model_io_plan.attach(resources=resources)
        if self._forward_hook_plan is not None:
            any_attached |= self._forward_hook_plan.attach(resources=resources)
        if self._backward_hook_plan is not None:
            any_attached |= self._backward_hook_plan.attach(resources=resources)
        return any_attached

    @classmethod
    def _build_model_io_plan(
        cls,
        context: ModelIOPlanBuildContext,
    ) -> Optional[ModelIOPlan[ModelInputTContr, ModelOutputTContr]]:
        plan_class = cls._get_model_io_plan()
        if plan_class is not None:
            return plan_class.build(context=context)

    @classmethod
    def _build_forward_hook_plan(
        cls,
        context: ForwardHookPlanBuildContext,
    ) -> Optional[ForwardHookPlan[ModelTContr]]:
        plan_class = cls._get_forward_hook_plan()
        if plan_class is not None:
            return plan_class.build(context=context)

    @classmethod
    def _build_backward_hook_plan(
        cls,
        context: BackwardHookPlanBuildContext,
    ) -> Optional[BackwardHookPlan[ModelTContr]]:
        plan_class = cls._get_backward_hook_plan()
        if plan_class is not None:
            return plan_class.build(context=context)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Sequence, Type, TypeVar

from artifact_core.typing import ArtifactResult
from artifact_experiment.spi.plans import CallbackExecutionPlan, PlanBuildContext

from artifact_torch._base.components.callbacks.model_io import (
    ModelIOArrayCallback,
    ModelIOArrayCollectionCallback,
    ModelIOCallback,
    ModelIOPlotCallback,
    ModelIOPlotCollectionCallback,
    ModelIOScoreCallback,
    ModelIOScoreCollectionCallback,
)
from artifact_torch._base.components.handler_suites.model_io import ModelIOCallbackHandlerSuite
from artifact_torch._base.components.resources.model_io import ModelIOCallbackResources
from artifact_torch._base.model.io import ModelInput, ModelOutput


@dataclass(frozen=True)
class ModelIOPlanBuildContext(PlanBuildContext):
    pass


ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
ModelIOPlanT = TypeVar("ModelIOPlanT", bound="ModelIOPlan[Any, Any]")


class ModelIOPlan(
    CallbackExecutionPlan[
        ModelIOCallbackHandlerSuite[ModelInputTContr, ModelOutputTContr],
        ModelIOCallback[ModelInputTContr, ModelOutputTContr, ArtifactResult, Any],
        ModelIOCallbackResources[ModelOutputTContr],
        ModelIOPlanBuildContext,
    ],
    Generic[ModelInputTContr, ModelOutputTContr],
):
    @classmethod
    @abstractmethod
    def _get_score_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOScoreCallback[ModelInputTContr, ModelOutputTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOArrayCallback[ModelInputTContr, ModelOutputTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOPlotCallback[ModelInputTContr, ModelOutputTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_score_collection_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOScoreCollectionCallback[ModelInputTContr, ModelOutputTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_collection_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOArrayCollectionCallback[ModelInputTContr, ModelOutputTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_collection_callbacks(
        cls, context: ModelIOPlanBuildContext
    ) -> Sequence[ModelIOPlotCollectionCallback[ModelInputTContr, ModelOutputTContr]]: ...

    def attach(self, resources: ModelIOCallbackResources[ModelOutputTContr]) -> bool:
        return self._handler_suite.attach(resources=resources)

    @classmethod
    def _get_handler_suite(
        cls,
    ) -> Type[ModelIOCallbackHandlerSuite[ModelInputTContr, ModelOutputTContr]]:
        return ModelIOCallbackHandlerSuite



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Mapping, Optional, Type, TypeVar

import pandas as pd
from artifact_core.table_comparison._artifacts.base import TableComparisonArtifactResources
from artifact_core.table_comparison.spi import TabularDataSpecProtocol
from artifact_experiment.table_comparison import TableComparisonPlan
from artifact_experiment.tracking import DataSplit

from artifact_torch._base.components.routines.artifact import (
    ArtifactRoutine,
    ArtifactRoutineData,
    ArtifactRoutineHyperparams,
)
from artifact_torch._domains.generation.model import GenerationParams
from artifact_torch.table_comparison._model import TableSynthesizer

GenerationParamsTCov = TypeVar("GenerationParamsTCov", bound=GenerationParams, covariant=True)
TableComparisonRoutineT = TypeVar("TableComparisonRoutineT", bound="TableComparisonRoutine")


@dataclass
class TableComparisonRoutineHyperparams(ArtifactRoutineHyperparams, Generic[GenerationParamsTCov]):
    generation_params: GenerationParamsTCov


@dataclass
class TableComparisonRoutineData(ArtifactRoutineData):
    df_real: pd.DataFrame


class TableComparisonRoutine(
    ArtifactRoutine[
        TableSynthesizer[Any, Any, GenerationParamsTCov],
        TableComparisonRoutineData,
        TableComparisonRoutineHyperparams[GenerationParamsTCov],
        TableComparisonArtifactResources,
        TabularDataSpecProtocol,
    ],
    Generic[GenerationParamsTCov],
):
    @classmethod
    @abstractmethod
    def _get_period(
        cls,
        data_split: DataSplit,
    ) -> Optional[int]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_plan(cls, data_split: DataSplit) -> Optional[Type[TableComparisonPlan]]: ...

    @classmethod
    @abstractmethod
    def _get_generation_params(cls) -> GenerationParamsTCov: ...

    @classmethod
    def _get_hyperparams(cls) -> TableComparisonRoutineHyperparams[GenerationParamsTCov]:
        generation_params = cls._get_generation_params()
        hyperparams = TableComparisonRoutineHyperparams[GenerationParamsTCov](
            generation_params=generation_params
        )
        return hyperparams

    def _generate_artifact_resources(
        self,
        model: TableSynthesizer[Any, Any, GenerationParamsTCov],
    ) -> Mapping[DataSplit, TableComparisonArtifactResources]:
        df_synthetic = model.generate(params=self._hyperparams.generation_params)
        resources_by_split = {}
        for data_split in self._data.keys():
            resources = TableComparisonArtifactResources(
                dataset_real=self._data[data_split].df_real, dataset_synthetic=df_synthetic
            )
            resources_by_split[data_split] = resources
        return resources_by_split



from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import torch
from artifact_core.shared.plotters import PDFConfig, PDFPlotter
from artifact_core.typing import Plot
from artifact_torch._base.components.callbacks.forward_hook import ForwardHookPlotCallback
from artifact_torch._base.model.base import Model
from artifact_torch._utils.collections.tensor_flattener import TensorFlattener
from torch import Tensor
from torch.nn import Module


class AllActivationsPDF(ForwardHookPlotCallback[Model[Any, Any]]):
    _base_key = "ACTIVATIONS_PDF"
    _hook_result_key = "layer_activations"
    _plot_feature_name = "All Layers — Activations"
    _max_activation_samples = 1_000_000

    @classmethod
    def _get_base_key(cls) -> str:
        return cls._base_key

    @classmethod
    def _get_layers(cls, model: Module) -> List[Module]:
        return [m for m in model.modules() if not any(m.children())]

    @classmethod
    def _hook(
        cls, module: Module, inputs: Tuple[Any, ...], output: Any
    ) -> Optional[Dict[str, Tensor]]:
        _ = module
        _ = inputs
        ls_t_out: List[Tensor] = []
        for t in TensorFlattener.flatten_tensors(x=output):
            ls_t_out.append(t.detach().reshape(-1).cpu())
        if ls_t_out:
            t_layer_activations = torch.cat(ls_t_out, dim=0)
            return {cls._hook_result_key: t_layer_activations}

    @classmethod
    def _aggregate(
        cls,
        hook_results: Dict[str, List[Dict[str, Tensor]]],
    ) -> Plot:
        collected = cls._collect_activations(hook_results)
        sr = cls._activations_to_series(collected)
        return cls._plot_pdf(sr)

    @classmethod
    def _collect_activations(
        cls,
        hook_results: Dict[str, List[Dict[str, Tensor]]],
    ) -> List[Tensor]:
        collected: List[Tensor] = []
        for ls_activations in hook_results.values():
            for activation in ls_activations:
                vals = activation.get(cls._hook_result_key, None)
                if vals is not None and vals.numel() > 0:
                    collected.append(vals)
        return collected

    @classmethod
    def _sample_activations(cls, t_activations: Tensor) -> Tensor:
        n_activations = t_activations.numel()
        if n_activations <= cls._max_activation_samples:
            return t_activations
        idx = torch.randperm(n_activations)[: cls._max_activation_samples]
        return t_activations.index_select(0, idx)

    @classmethod
    def _activations_to_series(cls, collected: List[Tensor]) -> pd.Series:
        if not collected:
            return pd.Series(dtype=float)
        t_all = torch.cat(collected, dim=0)
        t_all = cls._sample_activations(t_all)
        all_vals = t_all.numpy()
        return pd.Series(all_vals, dtype=float)

    @classmethod
    def _plot_pdf(cls, sr: pd.Series) -> Plot:
        return PDFPlotter.plot_pdf(
            sr_data=sr,
            feature_name=cls._plot_feature_name,
            config=PDFConfig(),
        )



from artifact_torch._base.components.callbacks.forward_hook import (
    ForwardHookArrayCallback,
    ForwardHookArrayCollectionCallback,
    ForwardHookCallback,
    ForwardHookPlotCallback,
    ForwardHookPlotCollectionCallback,
    ForwardHookScoreCallback,
    ForwardHookScoreCollectionCallback,
)
from artifact_torch._impl.callbacks.forward_hook.activation_pdf import AllActivationsPDF

__all__ = [
    "ForwardHookArrayCallback",
    "ForwardHookArrayCollectionCallback",
    "ForwardHookCallback",
    "ForwardHookPlotCallback",
    "ForwardHookPlotCollectionCallback",
    "ForwardHookScoreCallback",
    "ForwardHookScoreCollectionCallback",
    "AllActivationsPDF",
]



from artifact_torch._base.components.callbacks.backward_hook import (
    BackwardHookArrayCallback,
    BackwardHookArrayCollectionCallback,
    BackwardHookCallback,
    BackwardHookPlotCallback,
    BackwardHookPlotCollectionCallback,
    BackwardHookScoreCallback,
    BackwardHookScoreCollectionCallback,
)
from artifact_torch._base.components.callbacks.forward_hook import (
    ForwardHookArrayCallback,
    ForwardHookArrayCollectionCallback,
    ForwardHookCallback,
    ForwardHookPlotCallback,
    ForwardHookPlotCollectionCallback,
    ForwardHookScoreCallback,
    ForwardHookScoreCollectionCallback,
)
from artifact_torch._base.components.callbacks.model_io import (
    ModelIOArrayCallback,
    ModelIOArrayCollectionCallback,
    ModelIOCallback,
    ModelIOPlotCallback,
    ModelIOPlotCollectionCallback,
    ModelIOScoreCallback,
    ModelIOScoreCollectionCallback,
)
from artifact_torch._impl.callbacks.forward_hook.activation_pdf import AllActivationsPDF
from artifact_torch._impl.callbacks.model_io.loss import LossCallback

__all__ = [
    "BackwardHookArrayCallback",
    "BackwardHookArrayCollectionCallback",
    "BackwardHookCallback",
    "BackwardHookPlotCallback",
    "BackwardHookPlotCollectionCallback",
    "BackwardHookScoreCallback",
    "BackwardHookScoreCollectionCallback",
    "ForwardHookArrayCallback",
    "ForwardHookArrayCollectionCallback",
    "ForwardHookCallback",
    "ForwardHookPlotCallback",
    "ForwardHookPlotCollectionCallback",
    "ForwardHookScoreCallback",
    "ForwardHookScoreCollectionCallback",
    "ModelIOArrayCallback",
    "ModelIOArrayCollectionCallback",
    "ModelIOCallback",
    "ModelIOPlotCallback",
    "ModelIOPlotCollectionCallback",
    "ModelIOScoreCallback",
    "ModelIOScoreCollectionCallback",
    "AllActivationsPDF",
    "LossCallback",
]



from dataclasses import dataclass
from typing import Optional

import numpy as np


@dataclass
class TrackingScoreInfo:
    score_key: str
    latest_score: float
    best_score: float
    best_epoch: int


class TrainingProgressLogger:
    @staticmethod
    def get_progress_update(
        n_epochs_elapsed: int,
        train_loss: Optional[float] = None,
        val_loss: Optional[float] = None,
        tracking_score_info: Optional[TrackingScoreInfo] = None,
    ) -> str:
        desc = f"(Epoch {n_epochs_elapsed})"
        if train_loss is not None:
            desc += f"---> Train Loss: {train_loss:.4f}"
        if val_loss is not None:
            desc += f" | Val Loss: {val_loss:.4f}"
        if tracking_score_info is not None:
            if tracking_score_info.latest_score is not None and not np.isnan(
                tracking_score_info.latest_score
            ):
                score_recap = TrainingProgressLogger._get_score_recap(
                    score_key=tracking_score_info.score_key,
                    best_score=tracking_score_info.best_score,
                    best_epoch=tracking_score_info.best_epoch,
                )
                desc += (
                    f" |{tracking_score_info.score_key}: {tracking_score_info.latest_score:.4f}"
                    f" | {score_recap}"
                )
        return desc.strip()

    @staticmethod
    def _get_score_recap(score_key: str, best_score: float, best_epoch: int) -> str:
        return f"Best {score_key}: {best_score:.4f} @ (Epoch {best_epoch})"



from abc import abstractmethod
from typing import Any, Dict, Generic, Optional, Type, TypeVar

import pandas as pd
import torch
from artifact_experiment.tracking.spi import FileWriter
from torch import optim

from artifact_torch._base.components.cache.score_cache import ScoreCache
from artifact_torch._base.components.callbacks.checkpoint import CheckpointCallback
from artifact_torch._base.components.early_stopping.stopper import EarlyStopper, StopperUpdateData
from artifact_torch._base.components.logging.progress import TrainingProgressLogger
from artifact_torch._base.components.model_tracking.model_tracker import (
    ModelTracker,
    ModelTrackingCriterion,
)
from artifact_torch._base.components.resources.checkpoint import CheckpointCallbackResources
from artifact_torch._base.components.routines.artifact import ArtifactRoutine
from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine
from artifact_torch._base.data.data_loader import DataLoader
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.base import TrainerBase
from artifact_torch._base.trainer.routine_suite import RoutineSuite
from artifact_torch._base.trainer.training_state import TrainingState

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
StopperUpdateDataT = TypeVar("StopperUpdateDataT", bound=StopperUpdateData)
ModelTrackingCriterionT = TypeVar("ModelTrackingCriterionT", bound=ModelTrackingCriterion)
TrainerT = TypeVar("TrainerT", bound="Trainer")


class Trainer(
    TrainerBase[ModelTContr, ModelInputTContr, ModelOutputTContr],
    Generic[
        ModelTContr,
        ModelInputTContr,
        ModelOutputTContr,
        StopperUpdateDataT,
        ModelTrackingCriterionT,
    ],
):
    _train_loss_key = "LOSS_EPOCH"
    _val_loss_key = "LOSS_VAL"

    def __init__(
        self,
        training_state: TrainingState[ModelTContr],
        train_loader: DataLoader[ModelInputTContr],
        device: torch.device,
        early_stopper: EarlyStopper[StopperUpdateDataT],
        model_tracker: Optional[ModelTracker[ModelTrackingCriterionT]],
        checkpoint_callback: Optional[CheckpointCallback],
        routine_suite: RoutineSuite[ModelTContr, ModelInputTContr, ModelOutputTContr],
    ):
        super().__init__(training_state=training_state, train_loader=train_loader, device=device)
        self._early_stopper = early_stopper
        self._model_tracker = model_tracker
        self._checkpoint_callback = checkpoint_callback
        self._routine_suite = routine_suite
        self._epoch_score_cache = ScoreCache()

    @classmethod
    def build(
        cls: Type[TrainerT],
        model: ModelTContr,
        train_loader: DataLoader[ModelInputTContr],
        train_diagnostics_routine: Optional[
            TrainDiagnosticsRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ] = None,
        loader_routine: Optional[
            DataLoaderRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ] = None,
        artifact_routine: Optional[ArtifactRoutine[ModelTContr, Any, Any, Any, Any]] = None,
        file_writer: Optional[FileWriter] = None,
    ) -> TrainerT:
        optimizer = cls._get_optimizer(model=model)
        scheduler = cls._get_scheduler(optimizer=optimizer)
        training_state = TrainingState[ModelTContr](
            model=model,
            optimizer=optimizer,
            scheduler=scheduler,
        )
        device = cls._get_device()
        early_stopper = cls._get_early_stopper()
        model_tracker = cls._get_model_tracker()
        checkpoint_callback = cls._get_checkpoint_callback(writer=file_writer)
        routine_suite = RoutineSuite(
            train_diagnostics_routine=train_diagnostics_routine,
            artifact_routine=artifact_routine,
            loader_routine=loader_routine,
        )
        trainer = cls(
            training_state=training_state,
            train_loader=train_loader,
            device=device,
            early_stopper=early_stopper,
            model_tracker=model_tracker,
            checkpoint_callback=checkpoint_callback,
            routine_suite=routine_suite,
        )
        return trainer

    @property
    def epoch_scores(self) -> pd.DataFrame:
        return self._epoch_score_cache.scores

    @property
    def latest_model_state(self) -> Dict[str, Any]:
        return self._training_state.model.state_dict()

    @property
    def best_model_state(self) -> Optional[Dict[str, Any]]:
        if self._model_tracker is not None:
            return self._model_tracker.best_model_state

    @staticmethod
    @abstractmethod
    def _get_optimizer(model: ModelTContr) -> optim.Optimizer: ...

    @staticmethod
    @abstractmethod
    def _get_scheduler(
        optimizer: optim.Optimizer,
    ) -> Optional[optim.lr_scheduler._LRScheduler]: ...

    @staticmethod
    @abstractmethod
    def _get_device() -> torch.device: ...

    @staticmethod
    @abstractmethod
    def _get_checkpoint_period() -> Optional[int]: ...

    @staticmethod
    @abstractmethod
    def _get_early_stopper() -> EarlyStopper[StopperUpdateDataT]: ...

    @abstractmethod
    def _get_stopper_update_data(self) -> StopperUpdateDataT: ...

    @staticmethod
    @abstractmethod
    def _get_model_tracker() -> Optional[ModelTracker[ModelTrackingCriterionT]]: ...

    @abstractmethod
    def _get_model_tracking_criterion(self) -> Optional[ModelTrackingCriterionT]: ...

    @classmethod
    def _get_checkpoint_callback(
        cls,
        writer: Optional[FileWriter],
    ) -> Optional[CheckpointCallback]:
        period = cls._get_checkpoint_period()
        if period is not None and writer is not None:
            return CheckpointCallback(writer=writer, period=period)

    def _should_stop(self) -> bool:
        return self._early_stopper.stopped

    def _execute_epoch_preprocessing(self):
        self._routine_suite.attach_train_diagnostics_hooks(
            model=self._training_state.model, n_epochs_elapsed=self._n_epochs_elapsed
        )

    def _execute_epoch_postprocessing(self):
        super()._execute_epoch_postprocessing()
        self._execute_routine_suite()
        self._execute_checkpoint_callback()
        self._update_tracker()
        self._update_stopper()

    def _execute_routine_suite(self):
        self._routine_suite.clear_cache()
        self._routine_suite.execute(
            model=self._training_state.model, n_epochs_elapsed=self.n_epochs_elapsed
        )
        self._epoch_score_cache.append(self._routine_suite.scores)

    def _execute_checkpoint_callback(self):
        if self._checkpoint_callback is not None:
            checkpoint = self._get_checkpoint()
            checkpoint_callback_resources = CheckpointCallbackResources(
                export_data=checkpoint, epoch=self._n_epochs_elapsed
            )
            self._checkpoint_callback.execute(resources=checkpoint_callback_resources)

    def _update_tracker(self):
        if self._model_tracker is not None:
            criterion = self._get_model_tracking_criterion()
            assert criterion is not None
            self._model_tracker.update(
                model=self._training_state.model, criterion=criterion, epoch=self.n_epochs_elapsed
            )

    def _update_stopper(self):
        update_data = self._get_stopper_update_data()
        self._early_stopper.update(update_data=update_data)

    def _get_progress_bar_description(self) -> str:
        train_loss = self._epoch_score_cache.get_latest_non_null(key=self._train_loss_key)
        val_loss = self._epoch_score_cache.get_latest_non_null(key=self._val_loss_key)
        desc = TrainingProgressLogger.get_progress_update(
            n_epochs_elapsed=self.n_epochs_elapsed, train_loss=train_loss, val_loss=val_loss
        )
        return desc



from artifact_torch._base.components.early_stopping.single_score import SingleScoreStopper


class ScoreMinimizationStopper(SingleScoreStopper):
    def stopping_condition_met(self) -> bool:
        max_epochs_exceeded = self._max_epochs_exceeded()
        queue_is_full = self._queue_is_full()
        return max_epochs_exceeded or (queue_is_full and self.min_score == self.earliest_score)


class ScoreMaximizationStopper(SingleScoreStopper):
    def stopping_condition_met(self) -> bool:
        max_epochs_exceeded = self._max_epochs_exceeded()
        queue_is_full = self._queue_is_full()
        return max_epochs_exceeded or (queue_is_full and self.max_score == self.earliest_score)



from abc import abstractmethod
from typing import List, Optional

from artifact_torch._base.components.early_stopping.patience import PatienceStopper


class SingleScoreStopper(PatienceStopper[float]):
    @property
    def score_history(self) -> List[float]:
        return [criterion for criterion in self.criterion_history]

    @property
    def latest_score(self) -> Optional[float]:
        if self.latest_criterion is not None:
            return self.latest_criterion

    @property
    def earliest_score(self) -> Optional[float]:
        if self.earliest_criterion is not None:
            return self.earliest_criterion

    @property
    def min_score(self) -> Optional[float]:
        if not self._queue_is_empty():
            return min(self.score_history)

    @property
    def max_score(self) -> Optional[float]:
        if not self._queue_is_empty():
            return max(self.score_history)

    @abstractmethod
    def stopping_condition_met(self) -> bool:
        pass



from artifact_torch._base.components.model_tracking.model_tracker import (
    ModelTracker,
    ModelTrackingCriterion,
)
from artifact_torch._base.components.model_tracking.single_score import (
    SingleScoreCriterion,
    SingleScoreTracker,
)

__all__ = [
    "ModelTracker",
    "ModelTrackingCriterion",
    "SingleScoreCriterion",
    "SingleScoreTracker",
]



from abc import abstractmethod
from typing import Generic, TypeVar

from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._domains.classification.model import ClassificationParams, Classifier
from artifact_torch.binary_classification import BinaryClassificationResults

ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTCov = TypeVar("ModelOutputTCov", bound=ModelOutput, covariant=True)
ClassificationParamsTContr = TypeVar(
    "ClassificationParamsTContr", bound=ClassificationParams, contravariant=True
)
ClassificationDataTContr = TypeVar("ClassificationDataTContr", contravariant=True)


class BinaryClassifier(
    Classifier[
        ModelInputTContr,
        ModelOutputTCov,
        ClassificationParamsTContr,
        ClassificationDataTContr,
        BinaryClassificationResults,
    ],
    Generic[
        ModelInputTContr, ModelOutputTCov, ClassificationParamsTContr, ClassificationDataTContr
    ],
):
    @abstractmethod
    def forward(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov: ...

    @abstractmethod
    def classify(
        self, data: ClassificationDataTContr, params: ClassificationParamsTContr
    ) -> BinaryClassificationResults: ...



import torch
from artifact_experiment.spi.callbacks import ExportCallback
from artifact_experiment.tracking.spi import FileWriter

from artifact_torch._base.components.resources.checkpoint import CheckpointCallbackResources
from artifact_torch._utils.filesystem.filename_appender import FilenameModifier
from artifact_torch._utils.scheduling.periodic_actions import PeriodicActionTrigger


class CheckpointCallback(ExportCallback[CheckpointCallbackResources]):
    _export_name = "TRAINING_CHECKPOINTS"

    def __init__(self, period: int, writer: FileWriter):
        super().__init__(writer=writer)
        self._period = period

    @classmethod
    def _get_export_name(cls) -> str:
        return cls._export_name

    @classmethod
    def _get_extension(cls) -> str:
        return "pth"

    def execute(self, resources: CheckpointCallbackResources):
        if PeriodicActionTrigger.should_trigger(step=resources.epoch, period=self._period):
            super().execute(resources=resources)

    @classmethod
    def _save_local(cls, resources: CheckpointCallbackResources, filepath: str) -> str:
        filename = cls._get_filename(epoch=resources.epoch)
        filepath = FilenameModifier.replace(filepath=filepath, new_basename=filename).as_posix()
        torch.save(resources.export_data, filepath)
        return filepath

    @staticmethod
    def _get_filename(epoch: int) -> str:
        return f"EPOCH_{epoch}"



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

StopperUpdateDataTContr = TypeVar(
    "StopperUpdateDataTContr", bound="StopperUpdateData", contravariant=True
)


@dataclass
class StopperUpdateData:
    n_epochs_elapsed: int


class EarlyStopper(ABC, Generic[StopperUpdateDataTContr]):
    def __init__(self, max_n_epochs: Optional[int] = None):
        super().__init__()
        self._stopped: bool = False
        self._n_epochs_elapsed: int = 0
        self._max_n_epochs = max_n_epochs

    @property
    def stopped(self) -> bool:
        return self._stopped

    @property
    def n_epochs_elapsed(self) -> int:
        return self._n_epochs_elapsed

    @property
    def max_n_epochs(self) -> Optional[int]:
        return self._max_n_epochs

    @abstractmethod
    def stopping_condition_met(self) -> bool: ...

    def update(self, update_data: StopperUpdateDataTContr):
        self._update(update_data=update_data)
        self._stopped = self.stopping_condition_met()

    def _update(self, update_data: StopperUpdateDataTContr):
        self._n_epochs_elapsed = update_data.n_epochs_elapsed

    def _max_epochs_exceeded(self) -> bool:
        return self._max_n_epochs is not None and self._n_epochs_elapsed >= self._max_n_epochs



from typing import Dict, List

import torch
from artifact_torch._base.components.callbacks.model_io import ModelIOScoreCallback
from artifact_torch._base.model.io import ModelInput, ModelOutput


class LossCallback(ModelIOScoreCallback[ModelInput, ModelOutput]):
    _base_key = "LOSS"

    @classmethod
    def _get_base_key(cls):
        return cls._base_key

    @classmethod
    def _compute_on_batch(
        cls,
        model_input: ModelInput,
        model_output: ModelOutput,
    ) -> Dict[str, torch.Tensor]:
        _ = model_input
        t_loss = model_output.get("t_loss")
        assert t_loss is not None, "Loss tensor not provided"
        return {"t_loss": t_loss}

    @classmethod
    def _aggregate_batch_results(cls, ls_batch_results: List[Dict[str, torch.Tensor]]) -> float:
        ls_batch_losses = [result["t_loss"].item() for result in ls_batch_results]
        return sum(ls_batch_losses) / len(ls_batch_losses) if ls_batch_losses else float("nan")



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, TypeVar

import torch
import torch.nn as nn

from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound="ModelInput", contravariant=True)
ModelOutputTCov = TypeVar("ModelOutputTCov", bound="ModelOutput", covariant=True)


class Model(ABC, nn.Module, Generic[ModelInputTContr, ModelOutputTCov]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._device = torch.device("cpu")

    def __call__(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov:
        return super().__call__(model_input, *args, **kwargs)

    @property
    def params(self) -> Dict[str, Any]:
        return self.state_dict().copy()

    @property
    def param_count(self) -> Dict[str, int]:
        return self._get_param_count(model=self)

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, device: torch.device):
        self._move_to_device(device=device)

    @abstractmethod
    def forward(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov: ...

    def load_params(self, params: Dict[str, Any]):
        self.load_state_dict(state_dict=params)
        self._move_to_device(self._device)

    def _move_to_device(self, device: torch.device):
        self._device = device
        self.to(device)

    @staticmethod
    def _get_param_count(
        model: nn.Module,
    ) -> Dict[str, int]:
        total_params = sum(p.numel() for p in model.parameters())
        trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        non_trainable_params = total_params - trainable_params
        dict_param_count = {
            "total_parameters": total_params,
            "trainable_parameters": trainable_params,
            "non_trainable_parameters": non_trainable_params,
        }
        return dict_param_count



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Sequence, Type, TypeVar

from artifact_core.typing import ArtifactResult
from artifact_experiment.spi.plans import CallbackExecutionPlan, PlanBuildContext

from artifact_torch._base.components.callbacks.forward_hook import (
    ForwardHookArrayCallback,
    ForwardHookArrayCollectionCallback,
    ForwardHookCallback,
    ForwardHookPlotCallback,
    ForwardHookPlotCollectionCallback,
    ForwardHookScoreCallback,
    ForwardHookScoreCollectionCallback,
)
from artifact_torch._base.components.callbacks.hook import HookCallbackResources
from artifact_torch._base.components.handler_suites.hook import HookCallbackHandlerSuite
from artifact_torch._base.model.base import Model


@dataclass(frozen=True)
class ForwardHookPlanBuildContext(PlanBuildContext):
    pass


ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ForwardHookPlanT = TypeVar("ForwardHookPlanT", bound="ForwardHookPlan[Any]")


class ForwardHookPlan(
    CallbackExecutionPlan[
        HookCallbackHandlerSuite[ModelTContr],
        ForwardHookCallback[ModelTContr, ArtifactResult, Any],
        HookCallbackResources[ModelTContr],
        ForwardHookPlanBuildContext,
    ],
    Generic[ModelTContr],
):
    @classmethod
    @abstractmethod
    def _get_score_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookScoreCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookArrayCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookPlotCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_score_collection_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookScoreCollectionCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_array_collection_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookArrayCollectionCallback[ModelTContr]]: ...

    @classmethod
    @abstractmethod
    def _get_plot_collection_callbacks(
        cls, context: ForwardHookPlanBuildContext
    ) -> Sequence[ForwardHookPlotCollectionCallback[ModelTContr]]: ...

    def attach(self, resources: HookCallbackResources[ModelTContr]) -> bool:
        return self._handler_suite.attach(resources=resources)

    @classmethod
    def _get_handler_suite(cls) -> Type[HookCallbackHandlerSuite[ModelTContr]]:
        return HookCallbackHandlerSuite



from typing import Any, Iterable

import torch


class TensorFlattener:
    @classmethod
    def flatten_tensors(cls, x: Any) -> Iterable[torch.Tensor]:
        if torch.is_tensor(x):
            yield x
        elif isinstance(x, (list, tuple)):
            for y in x:
                yield from cls.flatten_tensors(y)
        elif isinstance(x, dict):
            for y in x.values():
                yield from cls.flatten_tensors(y)



from artifact_core.table_comparison import (
    TableComparisonArrayCollectionType,
    TableComparisonArrayType,
    TableComparisonPlotCollectionType,
    TableComparisonPlotType,
    TableComparisonScoreCollectionType,
    TableComparisonScoreType,
    TabularDataSpec,
)
from artifact_experiment.table_comparison._plan import TableComparisonPlan

from artifact_torch._domains.generation.model import GenerationParams
from artifact_torch.table_comparison._experiment import TabularSynthesisExperiment
from artifact_torch.table_comparison._model import TableSynthesizer
from artifact_torch.table_comparison._routine import (
    TableComparisonRoutine,
    TableComparisonRoutineData,
)

__all__ = [
    "TableComparisonArrayCollectionType",
    "TableComparisonArrayType",
    "TableComparisonPlotCollectionType",
    "TableComparisonPlotType",
    "TableComparisonScoreCollectionType",
    "TableComparisonScoreType",
    "TabularDataSpec",
    "TableComparisonPlan",
    "GenerationParams",
    "TabularSynthesisExperiment",
    "TableSynthesizer",
    "TableComparisonRoutine",
    "TableComparisonRoutineData",
]
